.well {
    max-width: 400px;
}

/* based on https://stackoverflow.com/a/30338814/1974918*/
details pre {
    padding: 0;
}

pre {
    overflow: auto;
    white-space: pre;
    word-wrap: normal;
    background-color: #ffffff;
}

code,
pre code {
    overflow: auto;
    white-space: pre;
    word-wrap: normal;
}

#copy,
#copy_data,
#copy_summary,
#copy_predict,
#copy_plot {
    position: absolute;
    right: 14px;
    font-size: 14px;
    color: grey;
    padding: 5px;
    opacity: 1;
}

#copy:hover,
#copy_data:hover,
#copy_summary:hover,
#copy_predict:hover,
#copy_plot:hover {
    color: white;
    opacity: 0.9;
}

#run:active {
    opacity: 0.5;
}